<CsoundSynthesizer>
<CsOptions>
-odac -dm3 --sample-accurate
</CsOptions>
<CsInstruments>

sr      = 44100
ksmps   = 100
nchnls  = 2

lua_opdef   "conform", {{
--lua_exec {{
package.path = package.path .. ";D:/Dropbox/silencio/?.lua;C:/Users/new/csound-csound6-git/?.lua;"
local ffi = require("ffi")
local math = require("math")
local string = require("string")
local Silencio = require ("Silencio")
local ChordSpace = require ("ChordSpace")
local csoundApi = ffi.load("csound64")
ffi.cdef[[
    int csoundGetKsmps(void *);
    double csoundGetSr(void *);
    int csoundInputMessage(void *, const char *message);
    int csoundMessage(void *, const char *format, ...);
]]
-- Print that we have acquired the global Csound object.
csoundApi.csoundMessage(csound, "csound: %p\\n", csound)
-- Prove that the Csound API and object are working.
csoundApi.csoundMessage(csound, "sr: %f\\n", csoundApi.csoundGetSr(csound))

ffi.cdef[[
    struct conform_t {
      double *noteout;
      double *notein;
      double *kin;
      double *qin;
      double *tin;
      double *iin;
      double k;
      double q;
      double t;
      double I;
      double kold;
      double qold;
      double told;
      double Iold;
    };
]]

local conform_ct = ffi.typeof("struct conform_t *")
csoundApi.csoundMessage(csound, "conform_ct: %p\\n", conform_ct)

local chord = Chord:new{0,2,4,7,11}
csoundApi.csoundMessage(csound, "chord: %s\\n", chord:information())


local function conform_init(csound, opcode, carguments)
    local p = ffi.cast(conform_ct, carguments)
    local pitch = p.notein[0]
    p.noteout[0] = pitch
    return 0
end

local function conform_kontrol(csound, opcode, carguments)
    local p = ffi.cast(conform_ct, carguments)
    return 0
end
}}


gasendL		init		0
gasendR		init		0 

            instr       FM

            ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
            ; Simple FM instrument.
            ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
            print       p1, p2, p3, p4, p5,  p6
khz         =           cpsmidinn(p4)
kamplitude  =           ampdb(p5) * 0.1
ktrackpady  chnget      "slider1"
kcarrier    =           1 + 10 * ktrackpady
ktrackpadx  chnget      "slider2"
kmodulator  =           1 + 10 * ktrackpadx
iattack     =           0.03
irelease    =           0.1
isustain    =           p3
p3          =           iattack + isustain + irelease
            ; Intensity sidebands.
intensity   chnget      "slider3"
aindex      transeg     intensity * 20, p3, -5, 0.25
isine       ftgenonce   1, 0, 16384, 10, 1
asignal     foscili     kamplitude, khz, kcarrier, kmodulator, k(aindex), isine
adamping    linseg      0, iattack, 1, isustain, 1, irelease, 0
asignal     =           asignal * adamping * aindex
gasendL     =           gasendL + asignal
gasendR     =           gasendR + asignal
            endin

		instr		reverb
print p1, p2, p3
kreverbdelay   init     .8
kreverbdelay_  chnget      "trackpad.x"
kreverbdelay   =       (kreverbdelay_ > 0? kreverbdelay_: kreverbdelay)
aL,aR		reverbsc	gasendL, gasendR,kreverbdelay, 12000
		outs		aL, aR
		clear		gasendL, gasendR
		endin

            instr       logistic
keyin       init        0
key         init        0
kk          init        0
kq          init        0
kt          init        0
ki          init        0
kc          init        .89
ky          init        .5
ky1         init        .5
ktempo      init        4
kgross      chnget      "slider4"
kgross      scale       kgross, .75, .999
kfine       chnget      "slider5"
kfine       scale       kfine, 0, .001
kc          =           kgross + kfine
ktimes      times
ktrigger    metro       ktempo
            if ktrigger == 1 then
ky1         = ky
ky          = kc * ky1 * (1 - ky1) * 4
keyin         = floor(36 + 60 * ky)
            lua_iopcall "chord", key, keyin,kk, kq, kt, ki
            event "i", 1, 0, 1/ktempo*4, key, 60
            endif
            printf      "nps:%9.4f t:%9.4f c:%12.8f y:%12.8f k:%4d\n", ktrigger, ktempo, ktimes, kc, ky, key

            
            endin
            
alwayson "reverb"
alwayson "logistic"

</CsInstruments>
<CsScore>
f 0 3600
</CsScore>
</CsoundSynthesizer>
